import sys

from PyQt5.QtGui import *
from qgis.analysis import QgsNativeAlgorithms
from qgis.core import *

# Tell Python where you will get processing from

sys.path.append(r'C:\Program Files\QGIS 3.16\apps\qgis-ltr\python\plugins')
# Reference the algorithm you want to run
from plugins import processing

from plugins.processing.algs.qgis.PointDistance import *

from plugins.processing.algs.qgis.HubDistanceLines import HubDistanceLines


def main():
    from qgis.core import QgsApplication

    # Start new Qgis application
    app = QGuiApplication([])
    QgsApplication.setPrefixPath(r'C:\Program Files\QGIS 3.0\apps\qgis', True)
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    QgsApplication.initQgis()
    feedback = QgsProcessingFeedback()

    # Some postprocessing for context
    def upload_new_layer(path, name):
        """Upload shp layers"""
        layer_name = "layer" + name
        provider_name = "ogr"
        layer = QgsVectorLayer(
            path,
            layer_name,
            provider_name)
        return layer

        ##########Create line that connect between each POI to nearest points along roads

    input_path = os.path.join(os.path.dirname(__file__), 'cliped.shp')
    input_clip = upload_new_layer(input_path, "input")
    hubs_path = os.path.join(os.path.dirname(__file__), 'points_along.shp')
    output_hub_dis_line = os.path.join(os.path.dirname(__file__), r'results_file/line_to_points.shp')

    alg = HubDistanceLines()
    alg.initAlgorithm()
    # Some preprocessing for context
    project = QgsProject.instance()

    target_crs = QgsCoordinateReferenceSystem()
    target_crs.createFromOgcWmsCrs(input_clip.crs().authid())
    project.setCrs(target_crs)
    context = QgsProcessingContext()
    context.setProject(project)
    params = {'INPUT': input_path, 'HUBS': hubs_path, 'FIELD': 'angle', 'UNIT': 4, 'OUTPUT': output_hub_dis_line}
    alg.processAlgorithm(params, context, feedback=feedback)

    ########## Buffer over constrain layer
    constrains = os.path.join(os.path.dirname(__file__), 'constrains.shp')
    buffer = os.path.join(os.path.dirname(__file__), "results_file/buffer.shp")

    params = {'INPUT': constrains, 'DISTANCE': 1, 'SEGMENTS': 1, 'END_CAP_STYLE': 2,
              'JOIN_STYLE': 1, 'MITER_LIMIT': 2,
              'DISSOLVE': True, 'OUTPUT': buffer}

    processing.run('native:buffer', params, feedback=feedback)

    ##########Convrt constrains layer as polylines
    constrains = buffer
    output_poly_as_lines = os.path.join(os.path.dirname(__file__), r'results_file/poly_as_lines.shp')
    params = {'INPUT': constrains, 'OUTPUT': output_poly_as_lines}
    # different implementation among QGIS versions
    processing.run("native:polygonstolines", params)

    # create line for other points in the list
    """For standalone application"""
    # Exit applications
    QgsApplication.exitQgis()
    app.exit()


if __name__ == "__main__":
    import cProfile, pstats
    from os import unlink, path

    # delete old sight line file
    file_path = 'test_same_area/sight_line'
    for file_ext in ('dbf', 'prj', 'shp', 'shx'):
        file_path_ext = '.'.join((file_path, file_ext))
        if path.isfile(file_path_ext):
            unlink(file_path_ext)

    profiler = cProfile.Profile()
    profiler.enable()
    main()
    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats('time')
    stats.print_stats()
